@@ -52,33 +52,40 @@
 #define RX_MTU 32767
 #define TX_MTU 32767
 
-#define TARGET_SIZE	16
-static const guint8 FTP_TARGET[TARGET_SIZE] = { 0xF9, 0xEC, 0x7B, 0xC4,
-					0x95, 0x3C, 0x11, 0xD2,
-					0x98, 0x4E, 0x52, 0x54,
-					0x00, 0xDC, 0x9E, 0x09 };
+#define TARGET_SIZE 16
+
+static const guint8 FTP_TARGET[TARGET_SIZE] = {
+			0xF9, 0xEC, 0x7B, 0xC4,  0x95, 0x3C, 0x11, 0xD2,
+			0x98, 0x4E, 0x52, 0x54,  0x00, 0xDC, 0x9E, 0x09  };
+
+static const guint8 PBAP_TARGET[TARGET_SIZE] = {
+			0x79, 0x61, 0x35, 0xF0,  0xF0, 0xC5, 0x11, 0xD8,
+			0x09, 0x66, 0x08, 0x00,  0x20, 0x0C, 0x9A, 0x66  };
 
 /* Connection ID */
 static guint32 cid = 0x0000;
 
 typedef struct {
-    guint8	version;
-    guint8	flags;
-    guint16	mtu;
+	guint8  version;
+	guint8  flags;
+	guint16 mtu;
 } __attribute__ ((packed)) obex_connect_hdr_t;
 
 struct obex_commands opp = {
 	.get		= opp_get,
-	.chkput		= opp_chkput,
 	.put		= opp_put,
-	.setpath	= NULL,
+	.chkput		= opp_chkput,
 };
 
 struct obex_commands ftp = {
 	.get		= ftp_get,
 	.put		= ftp_put,
-	.setpath	= ftp_setpath,
 	.chkput		= ftp_chkput,
+	.setpath	= ftp_setpath,
+};
+
+struct obex_commands pbap = {
+	.get		= pbap_get,
 };
 
 static void os_reset_session(struct obex_session *os)
@@ -207,19 +214,39 @@ static void cmd_connect(struct obex_session *os,
 	/* connection id will be used to track the sessions, even for OPP */
 	os->cid = ++cid;
 
-	if (os->target == NULL) {
-		register_transfer(os->cid, os);
-		/* OPP doesn't contains target or connection id. */
-		OBEX_ObjectSetRsp(obj, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
+	while (OBEX_ObjectGetNextHeader(obex, obj, &hi, &hd, &hlen)) {
+		if (hi != OBEX_HDR_TARGET || hlen != TARGET_SIZE)
+			continue;
+
+		if (memcmp(hd.bs, FTP_TARGET, TARGET_SIZE) == 0 &&
+				os->server->services & OBEX_FTP) {
+			os->target = FTP_TARGET;
+			os->cmds = &ftp;
+			break;
+		}
+
+		if (memcmp(hd.bs, PBAP_TARGET, TARGET_SIZE) == 0 &&
+				os->server->services & OBEX_PBAP) {
+			os->target = PBAP_TARGET;
+			os->cmds = &pbap;
+			break;
+		}
+
+		error("Connect attempt to a non-supported target");
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 		return;
 	}
 
-	hi = hlen = 0;
-	OBEX_ObjectGetNextHeader(obex, obj, &hi, &hd, &hlen);
+	if (os->target == NULL) {
+		if (os->server->services & OBEX_OPP) {
+			register_transfer(os->cid, os);
+			/* OPP doesn't contains target or connection id. */
+			OBEX_ObjectSetRsp(obj, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
+		} else {
+			error("Object Push connect attempt to a non-OPP server");
+			OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
+		}
 
-	if (hi != OBEX_HDR_TARGET || hlen != TARGET_SIZE
-			|| memcmp(os->target, hd.bs, TARGET_SIZE) != 0) {
-		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 		return;
 	}
 
@@ -227,6 +254,7 @@ static void cmd_connect(struct obex_session *os,
 	emit_session_created(cid);
 
 	/* Append received UUID in WHO header */
+	hd.bs = os->target;
 	OBEX_ObjectAddHeader(obex, obj,
 			OBEX_HDR_WHO, hd, TARGET_SIZE,
 			OBEX_FL_FIT_ONE_PACKET);
@@ -248,7 +276,7 @@ static gboolean chk_cid(obex_t *obex, obex_object_t *obj, guint32 cid)
 
 	os = OBEX_GetUserData(obex);
 
-	/* OPUSH doesn't provide a connection id. */
+	/* Object Push doesn't provide a connection id. */
 	if (os->target == NULL)
 		return TRUE;
 
@@ -274,7 +302,10 @@ static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 	guint hlen;
 	guint8 hi;
 
-	if (!os->cmds->get) {
+	if (!os->cmds) {
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
+		return;
+	} else if (!os->cmds->get) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
 				OBEX_RSP_NOT_IMPLEMENTED);
 		return;
@@ -348,7 +379,10 @@ static void cmd_setpath(struct obex_session *os,
 	guint32 hlen;
 	guint8 hi;
 
-	if (!os->cmds->setpath) {
+	if (!os->cmds) {
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
+		return;
+	} else if (!os->cmds->setpath) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
 				OBEX_RSP_NOT_IMPLEMENTED);
 		return;
@@ -652,7 +686,7 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 		return FALSE;
 	}
 
-	if (!os->cmds->chkput)
+	if (!os->cmds || !os->cmds->chkput)
 		goto done;
 
 	ret = os->cmds->chkput(obex, obj);
@@ -701,7 +735,10 @@ done:
 
 static void cmd_put(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 {
-	if (!os->cmds->put) {
+	if (!os->cmds) {
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
+		return;
+	} else if (!os->cmds->put) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
 				OBEX_RSP_NOT_IMPLEMENTED);
 		return;
@@ -728,7 +765,7 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 
 	switch (evt) {
 	case OBEX_EV_PROGRESS:
-		/* Just emit progress for OPUSH */
+		/* Just emit progress for Object Push */
 		if (os->target == NULL)
 			emit_transfer_progress(os->cid, os->size, os->offset);
 		break;
@@ -777,7 +814,7 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 	case OBEX_EV_REQCHECK:
 		switch (cmd) {
 		case OBEX_CMD_PUT:
-			if (os->cmds->put)
+			if (os->cmds && os->cmds->put)
 				check_put(obex, obj);
 			break;
 		default:
@@ -839,6 +876,15 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 	}
 }
 
+void server_free(struct server *server)
+{
+	g_free(server->name);
+	g_free(server->folder);
+	g_free(server->capability);
+	g_free(server->devnode);
+	g_free(server);
+}
+
 static void obex_handle_destroy(gpointer user_data)
 {
 	struct obex_session *os;
@@ -862,6 +908,17 @@ static void obex_handle_destroy(gpointer user_data)
 	OBEX_Cleanup(obex);
 }
 
+static gboolean tty_reinit(gpointer data)
+{
+	struct server *server = data;
+
+	tty_init(server->services, server->folder, server->capability, server->devnode);
+
+	server_free(server);
+
+	return FALSE;
+}
+
 static gboolean obex_handle_input(GIOChannel *io,
 				GIOCondition cond, gpointer user_data)
 {
@@ -870,8 +927,17 @@ static gboolean obex_handle_input(GIOChannel *io,
 	if (cond & G_IO_NVAL)
 		return FALSE;
 
-	if (cond & (G_IO_HUP | G_IO_ERR))
+	if (cond & (G_IO_HUP | G_IO_ERR)) {
+		struct obex_session *os;
+
+		error("HUP");
+
+		os = OBEX_GetUserData(obex);
+		if (os->server->devnode)
+			g_idle_add(tty_reinit, os->server);
+
 		return FALSE;
+	}
 
 	if (OBEX_HandleInput(obex, 1) < 0) {
 		error("Handle input error");
@@ -889,20 +955,11 @@ gint obex_session_start(gint fd, struct server *server)
 	gint ret;
 
 	os = g_new0(struct obex_session, 1);
-	switch (server->service) {
-	case OBEX_OPUSH:
-		os->target = NULL;
+
+	os->target = NULL;
+
+	if (server->services & OBEX_OPP)
 		os->cmds = &opp;
-		break;
-	case OBEX_FTP:
-		os->target = FTP_TARGET;
-		os->cmds = &ftp;
-		break;
-	default:
-		g_free(os);
-		debug("Invalid OBEX server");
-		return -EINVAL;
-	}
 
 	os->current_folder = g_strdup(server->folder);
 	os->server = server;
